#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2017-12-25
#
# Check_MK bgp_peers WATO plugin
#
# 2021-03-27: rewrite for CMK 2.0
# 2021-08-21: modified for bgp_peer plugin (from cisco_bgp_peer)
# 2021-08-29: removed htmloutput and infotext_values option
# 2022-04-02: added bgp neighbour states
# 2022-04-29: added upper/lower prefix limit
# 2022-05-09: added discovery rule set
# 2022-05-11: added remote_as to build_item
# 2022-09-05: added internal_item to avoid warnings on cmk updates (THX to Jay2k1 for reporting the issue)
# 2023-06-11: moved wato file from ~local/lib/check_mk/gui/plugins/wato
#             to ~/local/lib/check_mk/gui/plugins/wato/check_parameters to override the build in wato plugin
# 2024-06-15: fixed typo (no not)
#
# Known issue: the override of the build in wato plugin will break the build in aritsa bgp peer plugin
from click import help_option
from cmk.rulesets.v1 import Help, Label, Title
from cmk.rulesets.v1.form_specs import (
    DefaultValue,
    DictElement,
    Dictionary,
    InputHint,
    Integer,
    LevelDirection,
    LevelsType,
    List,
    MultipleChoice,
    MultipleChoiceElement,
    ServiceState,
    SimpleLevels,
    String,
    TimeMagnitude,
    TimeSpan,
    migrate_to_integer_simple_levels,
)
from cmk.rulesets.v1.rule_specs import CheckParameters, DiscoveryParameters, Topic, HostAndItemCondition
from cmk.rulesets.v1.form_specs.validators import NumberInRange, Message, LengthInRange

item_parts = [
    # MultipleChoiceElement(name='remote_address', title=Title('Peer remote address')),
    MultipleChoiceElement(name='remote_as', title=Title('Remote AS')),
    MultipleChoiceElement(name='address_family', title=Title('Address family')),
    MultipleChoiceElement(name='routing_instance', title=Title('Routing instance/VRF')),
]

def _migrate_to_float(value: object) -> float:
    if isinstance(value, int | float):
        return float(value)

    raise TypeError(value)

def _migrate_bgp_peer_list(value: object) -> list:
    if not isinstance(value, list):
        raise TypeError(value)

    peer_list = []

    for entry in value:
        if isinstance(entry, tuple):
            bgp_peer, bgp_peer_alias, state_not_found = entry
            peer_list.append({
                'bgp_peer': bgp_peer,
                'bgp_peer_alias': bgp_peer_alias,
                'state_not_found': state_not_found
            })
        else:
            peer_list.append(entry)

    return peer_list


def _migrate_neighbor_state(value: object) -> dict:
    if not isinstance(value, dict):
        raise TypeError(value)

    state_map = {}

    for state, value in value.items():
        match state:
            case '1':
                state_map['idle'] = value
            case '2':
                state_map['connect'] = value
            case '3':
                state_map['active'] = value
            case '4':
                state_map['open_sent'] = value
            case '5':
                state_map['open_confirm'] = value
            case '6':
                state_map['established'] = value
            case _:
                state_map[state] = value

    return state_map

def _parameter_form_check_bgp_peer():
    return Dictionary(
        elements={
            'minuptime': DictElement(
                parameter_form=SimpleLevels(
                    title=Title('Minimum uptime for peer'),
                    help_text=Help('Set the time in seconds, a peer must be up before the peer is considered sable.'),
                    prefill_levels_type=DefaultValue(LevelsType.FIXED),
                    prefill_fixed_levels=InputHint((7200, 3600)),
                    level_direction=LevelDirection.LOWER,
                    migrate=migrate_to_integer_simple_levels,
                    form_spec_template=TimeSpan(
                        displayed_magnitudes=[TimeMagnitude.SECOND, TimeMagnitude.MINUTE, TimeMagnitude.HOUR],
                        custom_validate=(NumberInRange(min_value=0),),
                        migrate=_migrate_to_float,
                    ))),
            'accepted_prefixes_upper_levels': DictElement(
                parameter_form=SimpleLevels(
                    title=Title('Accepted prefixes upper levels'),
                    help_text=Help('The values from WATO are preferred to the values from the device.'),
                    prefill_levels_type=DefaultValue(LevelsType.FIXED),
                    prefill_fixed_levels=InputHint((0, 0)),
                    level_direction=LevelDirection.UPPER,
                    migrate=migrate_to_integer_simple_levels,
                    form_spec_template=Integer(
                        unit_symbol='prefixes',
                        custom_validate=(NumberInRange(min_value=0, ),)
                    ))),
            'accepted_prefixes_lower_levels': DictElement(
                parameter_form=SimpleLevels(
                    title=Title('Accepted prefixes lower levels'),
                    help_text=Help('The values from WATO are preferred to the values from the device.'),
                    prefill_levels_type=DefaultValue(LevelsType.FIXED),
                    prefill_fixed_levels=InputHint((0, 0)),
                    level_direction=LevelDirection.LOWER,
                    migrate=migrate_to_integer_simple_levels,
                    form_spec_template=Integer(
                        unit_symbol='prefixes',
                        custom_validate=(NumberInRange(min_value=0, ),),
                    ))),
            'peernotfound': DictElement(
                parameter_form=ServiceState(
                    title=Title('State if peer is not found.'),
                    help_text=Help('Default monitoring state if the peer is not found in the SNMP data'),
                    prefill=DefaultValue(2)
                )),
            'admindown': DictElement(
                parameter_form=ServiceState(
                    prefill=DefaultValue(1),
                    title=Title('State if peer is admin shutdown.'),
                    help_text=Help('Monitoring state if the peer is admin shutdown')
                )),

            'neighborstate': DictElement(
                parameter_form=Dictionary(
                    title=Title('State to report for BGP neighbor state'),
                    help_text=Help('Map each BGP state to a CheckMK monitoring state'),
                    migrate=_migrate_neighbor_state,
                    elements={
                        'idle': DictElement(
                            parameter_form=ServiceState(
                                title=Title('1 - idle'),
                                help_text=Help(
                                    'This is the first stage of the BGP FSM. BGP detects a start event, tries to initiate a '
                                    'TCP connection to the BGP peer, and also listens for a new connect from a peer router. '
                                    'If an error causes BGP to go back to the Idle state for a second time, the '
                                    'ConnectRetryTimer is set to 60 seconds and must decrement to zero before the connection '
                                    'is initiated again. Further failures to leave the Idle state result in the '
                                    'ConnectRetryTimer doubling in length from the previous time. '
                                    'Default monitoring state is "CRIT"'
                                ),
                                prefill=DefaultValue(2)
                            )),
                        'connect': DictElement(
                            parameter_form=ServiceState(
                                title=Title('2 - connect'),
                                help_text=Help(
                                    'In this state, BGP initiates the TCP connection. If the 3-way TCP handshake completes, '
                                    'the established BGP Session BGP process resets the ConnectRetryTimer and sends the Open '
                                    'message to the neighbor, and then changes to the OpenSent State.'
                                    'Default monitoring state is "WARN"'),
                                prefill=DefaultValue(1),
                            )),
                        'active': DictElement(
                            parameter_form=ServiceState(
                                title=Title('3 - active'),
                                help_text=Help(
                                    'In this state, BGP starts a new 3-way TCP handshake. If a connection is established, '
                                    'an Open message is sent, the Hold Timer is set to 4 minutes, and the state moves to '
                                    'OpenSent. If this attempt for TCP connection fails, the state moves back to the Connect '
                                    'state and resets the ConnectRetryTimer. '
                                    'Default monitoring state is "WARN"'),
                                prefill=DefaultValue(1),
                            )),
                        'open_sent': DictElement(
                            parameter_form=ServiceState(
                                title=Title('4 - opensent'),
                                help_text=Help(
                                    'In this state, an Open message has been sent from the originating router and is awaiting '
                                    'an Open message from the other router. After the originating router receives the OPEN '
                                    'message from the other router, both OPEN messages are checked for errors. If the Open '
                                    'messages do not have any errors, the Hold Time is negotiated (using the lower value), '
                                    'and a KEEPALIVE message is sent (assuming the value is not set to zero). The connection '
                                    'state is then moved to OpenConfirm. If an error is found in the OPEN message, a '
                                    'Notification message is sent, and the state is moved back to Idle.'
                                    ' Default monitoring state is "WARN"'),
                                prefill=DefaultValue(1),
                            )),
                        'open_confirm': DictElement(
                            parameter_form=ServiceState(
                                title=Title('5 - openconfirm'),
                                help_text=Help(
                                    'In this state, BGP waits for a Keepalive or Notification message. Upon receipt of a '
                                    'neighbor’s Keepalive, the state is moved to Established. If the hold timer expires, a '
                                    'stop event occurs, or a Notification message is received, and the state is moved to '
                                    'Idle. '
                                    'Default monitoring state is "WARN"'),
                                prefill=DefaultValue(1),
                            )),
                        'established': DictElement(
                            parameter_form=ServiceState(
                                title=Title('6 - established'),
                                help_text=Help(
                                    'In this state, the BGP session is established. BGP neighbors exchange routes via Update '
                                    'messages. As Update and Keepalive messages are received, the Hold Timer is reset. If the '
                                    'Hold Timer expires, an error is detected and BGP moves the neighbor back to the Idle '
                                    'state. '
                                    'Default monitoring state is "OK"'),
                                prefill=DefaultValue(0),
                            )),
                    }
                )),

            'noprefixlimit': DictElement(
                parameter_form=ServiceState(
                    title=Title('State if no admin prefix limit/warn threshold is configured.'),
                    help_text=Help(
                        'The admin prefix limit and warn threshold needs to be configured on the device. '
                        'For example: "neighbor 172.17.10.10 maximum-prefix 10000 80". The threshold is in percentage '
                        'of the prefix limit.'
                    ),
                    prefill=DefaultValue(1),
                )),
            'peer_list': DictElement(
                parameter_form=List(
                    title=Title('BGP Peers'),
                    add_element_label=Label('Add BGP peer'),
                    remove_element_label=Label('remove BGP peer'),
                    no_element_label=Label('Add at least one BGP peer'),
                    migrate=_migrate_bgp_peer_list,
                    element_template=Dictionary(
                        elements={
                            'bgp_peer': DictElement(
                                required=True,
                                parameter_form=String(
                                    title=Title('BGP peer'),
                                    help_text=Help(
                                        'The configured value must match a BGP item reported by the monitored '
                                        'device. For example: "10.194.115.98" or "2A10:1CD0:1020:135::20 IPv6 Unicast"'
                                    ),
                                    custom_validate=(LengthInRange(min_value=1, ),),
                                )),
                            'bgp_peer_alias': DictElement(
                                parameter_form=String(
                                    title=Title('BGP Peer Alias'),
                                    help_text=Help(
                                        'You can configure an individual alias here for the BGP peer matching '
                                        'the text configured in the "BGP Peer IP-address" field. The alias will '
                                        'be shown in the check info'
                                    ),
                                    custom_validate=(LengthInRange(min_value=1, ),),
                                )),
                            'state_not_found': DictElement(
                                required=True,
                                parameter_form=ServiceState(
                                    title=Title('State if not found'),
                                    help_text=Help(
                                        'You can configure an individual state if the BGP peer matching the text '
                                        'configured in the "BGP Peer IP-address" field is not found'
                                    ),
                                    prefill=DefaultValue(2),
                                )),
                        }
                    )
                )),
            'internal_item': DictElement(
                render_only=True,
                parameter_form=MultipleChoice(
                    title=Title('Information not to use in the item name'),
                    elements=item_parts,
                )),
        }
    )


rule_spec_check_bgp_peer = CheckParameters(
    name='bgp_peer',
    parameter_form=_parameter_form_check_bgp_peer,
    title=Title('BGP Peer'),
    topic=Topic.NETWORKING,
    condition=HostAndItemCondition(item_title=Title('BGP peer')),
)


def _parameter_form_discovery_bgp_peer():
    return Dictionary(
        elements={
            'build_item': DictElement(
                parameter_form=MultipleChoice(
                    title=Title('Information not to use in the item name'),
                    help_text=Help(
                        'The Peer remote address is always used as the item name. By default the check will add the '
                        'address-family and the routing instance/VRF if available. You can decide to not use these '
                        'additional information in the item name. Do so only if your peers have only one address-'
                        'family configured and you don\'t have the same peer remote address in different routing '
                        'instances/VRFs configured.'
                    ),
                    elements=item_parts,
                    prefill=DefaultValue(['remote_as'])
                )
            )
        }
    )


rule_spec_discovery_bgp_beer = DiscoveryParameters(
    title=Title('BGP peer'),
    topic=Topic.NETWORKING,
    name='discovery_bgp_peer',
    parameter_form=_parameter_form_discovery_bgp_peer,
)
